from nbtex.LatexInterface.LatexFormatters import LatexBasicFormatter
from nbtex.core.operators import InvertibleOperator
from nbtex.core.Precedence import PRECEDENCE

f_combine = LatexBasicFormatter.function_definition
inv_f_combine = LatexBasicFormatter.inverse_function_definition

FunctionCall = InvertibleOperator(PRECEDENCE.POWER, f_combine, inv_f_combine)



from nbtex.core.operators.Operator import InvertibleOperator
from nbtex.core.Precedence import PRECEDENCE
from nbtex.LatexInterface.LatexFormatters import LatexBasicFormatter


def make_subscript(base, subs):
    return base + LatexBasicFormatter.subscript(subs)


Subscript = InvertibleOperator(PRECEDENCE.POWER, make_subscript)
Bold = InvertibleOperator(PRECEDENCE.POWER, LatexBasicFormatter.bold)
Italics = InvertibleOperator(PRECEDENCE.POWER, LatexBasicFormatter.italics)
Underline = InvertibleOperator(PRECEDENCE.POWER, LatexBasicFormatter.underline)



from IPython.display import Latex


def addEnv(s, env="gather*"):
    env = "{" + env + "}"
    return r"\begin" + env + str(s) + r"\end" + env


def latex(v):
    return Latex(addEnv(v.build()))


def platex(v, returnLatexObject=True):
    l = Latex(addEnv(v.build()))
    print(l.data)
    if returnLatexObject:
        return l



import unittest
import nbtex.core.operators.Arithmatic as A
import nbtex.core.operators.Derivative as D
import nbtex.core.operators.FunctionCall as F
import nbtex.core.operators.Series as S
import nbtex.core.operators.UnaryFormatting as U

# TODO: Writing tests for Derivative, FunctionCall, Series, UnaryFormatting
class TestArithmaticOperators(unittest.TestCase):
    def test_add_minus(self):
        plus = A.Plus("1", "2")
        minus = A.Minus("1", "2")
        self.assertEqual(plus, "1 + 2")
        self.assertEqual(minus, "1 - 2")
        self.assertEqual(plus, (~A.Minus)("1", "2"))
        self.assertEqual(minus, (~A.Plus)("1", "2"))

    def test_mult_frac(self):
        mult = A.Multiply("1", "2")
        frac = A.Fraction("1", "2")
        self.assertEqual(mult, "1 * 2")
        self.assertEqual(frac, r"\frac{1}{2}")
        self.assertEqual(mult, (~A.Fraction)("1", "2"))
        self.assertEqual(frac, (~A.Multiply)("1", "2"))

    def test_power_root(self):
        power = A.Power('1','2')
        root = A.Root('1', '2')
        self.assertEqual(power, '1 ^ 2')
        self.assertEqual(root, r'\sqrt[\leftroot{-1}\uproot{1}2]{1}')
        self.assertEqual(power, (~A.Root)('1','2'))
        self.assertEqual(root, (~A.Power)('1', '2'))

    def test_negate(self):
        negated = A.Negate('1')
        still_negated = A.Negate('-1')
        unnegated = (~A.Negate)('-1')
        still_unnegated = (~A.Negate)('1')
        self.assertEqual(negated, '-1')
        self.assertEqual(unnegated, '1')
        self.assertEqual(still_negated, '-1')
        self.assertEqual(still_unnegated, '1')

    def test_space(self):
        space = A.Space('1', '2')
        self.assertEqual(space, r'1 \hspace{1mm} 2')
    
    def test_compare_operators(self):
        # TODO: All relational operators tests
        pass




import nbtex.core.operators as Operator
from nbtex.helpers import varArgFunc
from nbtex.core.Precedence import PRECEDENCE
from nbtex.LatexInterface import LatexBasicFormatter
from nbtex.platforms.jupyter import latex


class Var:
    def __init__(self, name, backward_operation=None):
        self._name, self._backward_operation = name, backward_operation

    def clone(self):
        if self.backward_operation is None:
            return Var(self._name)
        return Var(self._name, self._backward_operation.clone())

    def __str__(self):
        return self.build()

    def __repr__(self):
        return self.__str__()

    @property
    def backward_operation(self):
        return self._backward_operation

    @property
    def precedence(self):
        return (
            PRECEDENCE.VAR
            if self._backward_operation is None
            else self._backward_operation.precedence
        )

    def build(self):
        return (
            self._name
            if self._backward_operation is None
            else self._backward_operation.build()
        )

    @staticmethod
    def operation(operator, *args):
        operation = Operation(operator)
        args = makeVar(*args)
        if isinstance(args, list):
            output = Var("output", operation(*args))
        else:
            output = Var("output", operation(args))
        return output

    def __invert__(self):
        if self._backward_operation is not None:
            self._backward_operation = self._backward_operation.__invert__()
        return self

    def __call__(self, *args):
        return Var.operation(Operator.FunctionCall, self, *args)

    def __add__(self, other):
        return Var.operation(Operator.Plus, self, other)

    def __sub__(self, other):
        return Var.operation(Operator.Minus, self, other)

    def __mul__(self, other):
        return Var.operation(Operator.Multiply, self, other)

    def __pow__(self, other):
        return Var.operation(Operator.Power, self, other)

    def __or__(self, other):
        return Var.operation(Operator.Space, self, other)

    def __and__(self, other):
        return Var.operation(Operator.Newline, self, other)

    def __lt__(self, other):
        return Var.operation(Operator.LessThan, self, other)

    def __gt__(self, other):
        return Var.operation(Operator.GreaterThan, self, other)

    def __le__(self, other):
        return Var.operation(Operator.LessThanEqual, self, other)

    def __ge__(self, other):
        return Var.operation(Operator.GreaterThanEqual, self, other)

    def __neg__(self):
        return Var.operation(Operator.Negate, self)

    def __truediv__(self, other):
        return Var.operation(Operator.Fraction, self, other)

    def equals(self, other):
        return Var.operation(Operator.Equal, self, other)

    def not_equals(self, other):
        return Var.operation(~Operator.Equal, self, other)

    def root(self, power):
        return Var.operation(Operator.Root, self, power)

    def sum_over(self, wrt="", lower="", upper=""):
        return Var.operation(Operator.Summation, self, wrt, lower, upper)

    def product_over(self, wrt="", lower="", upper=""):
        return Var.operation(Operator.Product, self, wrt, lower, upper)

    def integral_over(self, wrt="", lower="", upper=""):
        return Var.operation(Operator.Integral, self, wrt, lower, upper)

    def differentiate(self, wrt):
        return Var.operation(Operator.Derivative, self, wrt)

    def partial_integral_over(self, wrt="", lower="", upper=""):
        return Var.operation(Operator.PartialIntegral, self, wrt, lower, upper)

    def partial_differentiate(self, wrt):
        return Var.operation(Operator.PartialDerivative, self, wrt)

    def subscript(self, subs):
        return Var.operation(Operator.Subscript, self, subs)

    def bold(self):
        return self.use_operator_on_leaf(Operator.Bold)

    def italics(self):
        return self.use_operator_on_leaf(Operator.Italics)

    def underline(self):
        return self.use_operator_on_leaf(Operator.Underline)

    def use_operator_on_leaf(self, operator):
        if self.backward_operation is None:
            return Var.operation(operator, self)
        else:
            self._backward_operation = self.backward_operation.use_operator_on_leaf(
                operator
            )
            return self


def isVar(a):
    return isinstance(a, Var)


def makeVar(*args):
    return varArgFunc(lambda arg: arg.clone() if isVar(arg) else Var(str(arg)), *args)


def addEnv(s, env="gather*"):
    env = "{" + env + "}"
    return r"\begin" + env + str(s) + r"\end" + env


class Operation:
    def __init__(self, operator):
        self._operator = operator
        self._args = None

    def clone(self):
        return self
        # NOTE: Deep Clone: This is a costly operation
        #       Shallow cloning done for now
        # cloned_op = Operation(self._operator)
        # if self._args is None: return cloned_op
        # return cloned_op(*self._args)

    @property
    def precedence(self):
        return self._operator.precedence

    def __call__(self, *args):
        self._args = args
        return self

    def use_operator_on_leaf(self, operator):
        args = [arg.use_operator_on_leaf(operator) for arg in self._args]
        self._args = args
        return self

    def _is_invertible(self, op):
        return isinstance(op, Operator.InvertibleOperator)

    def build(self):
        args = [self._build_and_surround(makeVar(arg)) for arg in self._args]
        return self._operator(*args)

    def _build_and_surround(self, var_instance):
        return (
            self._get_surrounded_with_parens(var_instance.build())
            if self._should_surround_with_parens(var_instance)
            else var_instance.build()
        )

    def _get_surrounded_with_parens(self, string):
        return LatexBasicFormatter.surround_with_parens(string)

    def _should_surround_with_parens(self, var_instance):
        return (
            var_instance.backward_operation is not None
            and var_instance.backward_operation.precedence < self.precedence
        )

    def __invert__(self):
        if self._is_invertible(self._operator):
            self._operator = ~self._operator
        return self



from nbtex.core.Ast import Var, makeVar
from nbtex.Matrix import Matrix, MatrixBuilder, MatrixWithDots
from nbtex.Vector import Vector, makeVector
from nbtex.Dots import Dots
from nbtex.platforms.jupyter import latex, platex
from nbtex.Symbols import Greek, Set, Operator, Arrow


from nbtex.LatexInterface.LatexFormatters.LatexBasicFormatter import LatexBasicFormatter
from nbtex.LatexInterface.Operator import Operator


class LatexSeriesFormatter:
    @staticmethod
    def integral(func, wrt, lower="", upper=""):
        func, wrt, lower, upper = LatexSeriesFormatter._preprocess_args(
            func, f"d{wrt}", lower, upper
        )
        return f"\\int{lower}{upper} {func} {Operator.space} {wrt}"

    @staticmethod
    def partial_integral(func, wrt, lower="", upper=""):
        func, wrt, lower, upper = LatexSeriesFormatter._preprocess_args(
            func, f"\\partial {wrt}", lower, upper
        )
        return f"\\int{lower}{upper} {func} {Operator.space} {wrt}"

    @staticmethod
    def summation(func, wrt="", lower="", upper=""):
        lower = f"{wrt}={lower}" if lower != "" and wrt != "" else wrt
        func, _, lower, upper = LatexSeriesFormatter._preprocess_args(
            func, wrt, lower, upper
        )
        return f"\\sum{lower}{upper} {func}"

    @staticmethod
    def product(func, wrt="", lower="", upper=""):
        lower = f"{wrt}={lower}" if lower != "" and wrt != "" else wrt
        func, _, lower, upper = LatexSeriesFormatter._preprocess_args(
            func, wrt, lower, upper
        )
        return f"\\prod{lower}{upper} {func}"

    @staticmethod
    def _preprocess_args(func, wrt, lower, upper):
        lower, upper = LatexSeriesFormatter._format_lower_upper(lower, upper)
        func, wrt = LatexBasicFormatter.surround_all_with_braces(func, wrt)
        return func, wrt, lower, upper

    @staticmethod
    def _format_lower_upper(lower, upper):
        lower = LatexBasicFormatter.surround_with_braces(lower) if lower != "" else ""
        if lower != "":
            lower = LatexBasicFormatter.subscript(lower)

        upper = LatexBasicFormatter.surround_with_braces(upper) if upper != "" else ""
        if upper != "":
            upper = LatexBasicFormatter.superscript(upper)
        return lower, upper



from nbtex.core.Ast import makeVar
class SetSymbols:
    def __init__(self):
        pass

    @property
    def Subset(self):
        return makeVar(r"\Subset")

    @property
    def sqsubset(self):
        return makeVar(r"\sqsubset")

    @property
    def triangleleft(self):
        return makeVar(r"\triangleleft")

    @property
    def blacktriangleleft(self):
        return makeVar(r"\blacktriangleleft")

    @property
    def supset(self):
        return makeVar(r"\supset")

    @property
    def Supset(self):
        return makeVar(r"\Supset")

    @property
    def sqsupset(self):
        return makeVar(r"\sqsupset")

    @property
    def triangleright(self):
        return makeVar(r"\triangleright")

    @property
    def blacktriangleright(self):
        return makeVar(r"\blacktriangleright")

    @property
    def cap(self):
        return makeVar(r"\cap")

    @property
    def Cap(self):
        return makeVar(r"\Cap")

    @property
    def sqcap(self):
        return makeVar(r"\sqcap")

    @property
    def vartriangle(self):
        return makeVar(r"\vartriangle")

    @property
    def blacktriangle(self):
        return makeVar(r"\blacktriangle")

    @property
    def cup(self):
        return makeVar(r"\cup")

    @property
    def Cup(self):
        return makeVar(r"\Cup")

    @property
    def sqcup(self):
        return makeVar(r"\sqcup")

    @property
    def triangledown(self):
        return makeVar(r"\triangledown")

    @property
    def blacktriangledown(self):
        return makeVar(r"\blacktriangledown")

    @property
    def In(self):
        return makeVar(r"\in")

    @property
    def subseteq(self):
        return makeVar(r"\subseteq")

    @property
    def subseteqq(self):
        return makeVar(r"\subseteqq")

    @property
    def sqsubseteq(self):
        return makeVar(r"\sqsubseteq")

    @property
    def ni(self):
        return makeVar(r"\ni")

    @property
    def supseteq(self):
        return makeVar(r"\supseteq")

    @property
    def supseteqq(self):
        return makeVar(r"\supseteqq")

    @property
    def sqsupseteq(self):
        return makeVar(r"\sqsupseteq")

    @property
    def notin(self):
        return makeVar(r"\notin")

    @property
    def nsubseteq(self):
        return makeVar(r"\nsubseteq")

    @property
    def subsetneq(self):
        return makeVar(r"\subsetneq")

    @property
    def varsubsetneq(self):
        return makeVar(r"\varsubsetneq")

    @property
    def nsubseteqq(self):
        return makeVar(r"\nsubseteqq")

    @property
    def subsetneqq(self):
        return makeVar(r"\subsetneqq")

    @property
    def varsubsetneqq(self):
        return makeVar(r"\varsubsetneqq")

    @property
    def uplus(self):
        return makeVar(r"\uplus")

    @property
    def nsupseteq(self):
        return makeVar(r"\nsupseteq")

    @property
    def supsetneq(self):
        return makeVar(r"\supsetneq")

    @property
    def varsupsetneq(self):
        return makeVar(r"\varsupsetneq")

    @property
    def nsupseteqq(self):
        return makeVar(r"\nsupseteqq")

    @property
    def supsetneqq(self):
        return makeVar(r"\supsetneqq")

    @property
    def varsupsetneqq(self):
        return makeVar(r"\varsupsetneqq")


Set = SetSymbols()


from nbtex.LatexInterface.LatexFormatters import LatexBasicFormatter


class LatexDerivativesFormatter:
    @staticmethod
    def is_one_word(expr):
        for w in expr:
            if not w.isalnum() and w != "_" and w != "^":
                return False
        return True

    @staticmethod
    def partial(wrt):
        return LatexBasicFormatter.fraction("\\partial", f"\\partial {wrt}")

    @staticmethod
    def partial_of(function, wrt):
        if not LatexDerivativesFormatter.is_one_word(function):
            function = LatexBasicFormatter.surround_with_parens(function)
            par = LatexDerivativesFormatter.partial(wrt)
            return f"{par} {function}"
        return LatexBasicFormatter.fraction(f"\\partial {function}", f"\\partial {wrt}")

    @staticmethod
    def derivative(wrt):
        return LatexBasicFormatter.fraction("d", f"d{wrt}")

    @staticmethod
    def derivative_of(function, wrt):
        if not LatexDerivativesFormatter.is_one_word(function):
            function = LatexBasicFormatter.surround_with_parens(function)
            der = LatexDerivativesFormatter.derivative(wrt)
            return f"{der} {function}"
        return LatexBasicFormatter.fraction(f"d{function}", f"d{wrt}")



class OperatorSymbols:
    def __init__(self):
        pass

    @property
    def plus(self):
        return "+"

    @property
    def minus(self):
        return "-"

    @property
    def multiply(self):
        return "*"

    @property
    def power(self):
        return "^"
    
    @property
    def newline(self):
        return "\\\\"

    @property
    def space(self):
        return "\\hspace{1mm}"

    @property
    def eq(self):
        return "="

    @property
    def lt(self):
        return "<"

    @property
    def leq(self):
        return r"\leq"

    @property
    def gt(self):
        return ">"

    @property
    def geq(self):
        return r"\geq"

    @property
    def nless(self):
        return r"\nless"

    @property
    def nleq(self):
        return r"\nleq"

    @property
    def ngtr(self):
        return r"\ngtr"

    @property
    def ngeq(self):
        return r"\ngeq"

    @property
    def neq(self):
        return r"\neq"


Operator = OperatorSymbols()



from nbtex.core.Ast import Var, makeVar, isVar
from nbtex.LatexInterface.LatexFormatters import LatexMatrixFormatter
from nbtex.Dots import Dots

class Matrix(Var):
    def __init__(self, matrix=[[]], surround="[]"):
        super().__init__("matrix")
        self.matrix, self.surround = matrix, surround
        self.power, self.subs = None, None
    
    def clone(self):
        new_mtx = [[0]*len(self.matrix[0])]*len(self.matrix)
        for i in range(len(self.matrix)):
            for j in range(len(self.matrix[i])):
                if isVar(self.matrix[i][j]):
                    new_mtx[i][j] = self.matrix[i][j].clone()
                else:
                    new_mtx[i][j] = self.matrix[i][j]
        return Matrix(new_mtx, self.surround)

    def __getitem__(self, ind):
        if ind >= len(self.matrix):
            raise Exception("index out of bounds")
        if len(self.matrix) == 0:
            raise Exception("empty matrix")
        if isinstance(self.matrix[0], list):
            return Matrix(self.matrix[ind])
        else:
            return Var(self.matrix[ind])

    def __pow__(self, power):
        self.power = makeVar(power)
        return self

    def subscript(self, subs):
        self.subs = makeVar(subs)
        return self

    @staticmethod
    def builder():
        return MatrixBuilder()

    @property
    def T(self):
        if len(self.matrix) == 0:
            return self
        rows, cols = len(self.matrix), len(self.matrix[0])
        matrix = [[self.matrix[j][i] for j in range(rows)] for i in range(cols)]
        return Matrix(matrix, self.surround)

    def build(self):
        begin, end = self.create_env()
        mtx = self.get_built_base_matrix()
        # print(mtx)
        power, subs = self.get_pow_subs()
        return r"" + begin + mtx + end + power + subs + r""

    def create_env(self):
        return LatexMatrixFormatter.create_env(self.surround)

    def get_built_base_matrix(self):
        mtx = self.get_built_matrix_elements()
        return LatexMatrixFormatter.matrix_to_latex(mtx)

    def get_built_matrix_elements(self):
        mtx = [[makeVar(cell) for cell in row] for row in self.matrix]
        mtx = [[cell.build() for cell in row] for row in mtx]
        return mtx

    def get_pow_subs(self):
        if self.power is not None:
            power = LatexMatrixFormatter.power_of_matrix(self.power.build())
        else:
            power = LatexMatrixFormatter.empty_power()

        if self.subs is not None:
            subs = LatexMatrixFormatter.subscript_of_matrix(self.subs.build())
        else:
            subs = LatexMatrixFormatter.empty_subscript()
        return power, subs


class MatrixBuilder:
    def __init__(self):
        self.matrix, self.cols = [], None

    def add(self, *args):
        if self.cols is None:
            self.cols = len(args)
        elif not self.cols == len(args):
            raise Exception(
                f"Different dimensions for matrix builder. dim1:{self.cols} dim2:{len(args)}"
            )
        self.matrix.append(list(args))
        return self

    def create(self, surround="[]"):
        return Matrix(self.matrix, surround)


def hasHoriDots(elements, m):
    elerowlen = max([len(ele) for ele in elements])
    return not elerowlen == len(m[0])


def hasVertDots(elements, m):
    return not len(elements) == len(m)


class MatrixWithDots(Matrix):
    def __init__(self, elements, shape=None, surround="[]"):
        # resolve elements
        self._elements_check(elements)
        self.elements = elements
        rows, cols = self._get_dimensions(elements, shape)
        matrix = self._construct_matrix(rows, cols, elements)
        super().__init__(matrix, surround)

    def _elements_check(self, elements):
        if (
            not isinstance(elements, list)
            or len(elements) == 0
            or not isinstance(elements[0], list)
        ):
            raise Exception("need to pass list of list")

    def _get_dimensions(self, elements, shape):
        minrows, mincols = len(elements), max([len(row) for row in elements])
        if shape is None:
            rows = minrows + 1 if minrows != 1 else minrows
            cols = mincols + 1 if mincols != 1 else mincols
        else:
            try:
                rows, cols = shape
                rows, cols = int(rows), int(cols)
            except:
                raise Exception("shape should be of type tuple(int,int)")
            if rows < minrows or cols < mincols:
                raise Exception("shape >= (row, cols) of elements")
        return rows, cols

    def _construct_matrix(self, rows, cols, elements):
        m = self._get_zero_matrix_of_dimensions(rows, cols)
        i = self._fill_with_elements(0, elements, m)
        i = self._fill_with_vertical_dots(i, elements, m)
        self._fill_last_row_with_elements(elements, m)
        self._put_diagonal_dot(elements, m)
        return m

    def _get_zero_matrix_of_dimensions(self, rows, cols):
        return [[0] * cols for i in range(rows)]

    def _fill_with_elements(self, start, elements, m):
        i = start
        while i < len(elements) - 1:
            m[i] = self._get_row_with_elements(elements, i, len(m[i]))
            i += 1
        return i

    def _get_row_with_elements(self, elements, row_index, mlen):
        row = elements[row_index]
        new_row = [0] * mlen
        j = 0
        while j < len(row) - 1:
            new_row[j] = row[j]
            j += 1
        while j < mlen - 1:
            new_row[j] = Dots("h")
            j += 1
        new_row[j] = row[len(row) - 1]
        return new_row

    def _fill_with_vertical_dots(self, start, elements, m):
        i = start
        while i < len(m) - 1:
            m[i] = [Dots("v")] * len(m[i])
            i += 1
        
        return i

    def _fill_last_row_with_elements(self, elements, m):
        m[len(m) - 1] = self._get_row_with_elements(
            elements, len(elements) - 1, len(m[len(m) - 1])
        )

    def _put_diagonal_dot(self, elements, m):
        dli, dlj = len(m) - 2, len(m[0]) - 2
        if (
            isinstance(m[dli][dlj], Dots)
            and hasHoriDots(elements, m)
            and hasVertDots(elements, m)
        ):

            m[dli][dlj] = Dots("d")

    @property
    def T(self):
        raise Exception("transpose of dotted matrix not allowed")



import unittest
from nbtex.LatexInterface.LatexFormatters import LatexMatrixFormatter


class TestLatexMatrixFormatter(unittest.TestCase):
    def test_matrix_to_latex(self):
        mtx = [["1", "2", "3"], ["4", "5", "6"]]
        ltx_mtx = LatexMatrixFormatter.matrix_to_latex(mtx)
        self.assertEqual(ltx_mtx, r" 1 &  2 &  3 \\ 4 &  5 &  6 ")

    def test_empty_power_subscript(self):
        emppow = LatexMatrixFormatter.empty_power()
        empsub = LatexMatrixFormatter.empty_subscript()
        subscript = LatexMatrixFormatter.subscript_of_matrix('1')
        power = LatexMatrixFormatter.power_of_matrix('1')
        self.assertEqual(emppow, "")
        self.assertEqual(empsub, "")
        self.assertEqual(subscript, r"_{1}")
        self.assertEqual(power, r"^{1}")

    def test_create_env(self):
        beginb, endb = LatexMatrixFormatter.create_env("[]")
        beginp, endp = LatexMatrixFormatter.create_env("()")
        beginv, endv = LatexMatrixFormatter.create_env("||")
        beginV, endV = LatexMatrixFormatter.create_env("||||")
        begin, end = LatexMatrixFormatter.create_env("else")

        self.assertEqual(beginb, r"\begin{bmatrix}")
        self.assertEqual(endb, r"\end{bmatrix}")
        self.assertEqual(beginp, r"\begin{pmatrix}")
        self.assertEqual(endp, r"\end{pmatrix}")
        self.assertEqual(beginv, r"\begin{vmatrix}")
        self.assertEqual(endv, r"\end{vmatrix}")
        self.assertEqual(beginV, r"\begin{Vmatrix}")
        self.assertEqual(endV, r"\end{Vmatrix}")
        self.assertEqual(begin, r"\begin{matrix}")
        self.assertEqual(end, r"\end{matrix}")




class PrecedenceValues:
    def __init__(self):
        pass

    @property
    def POWER(self):
        return 13

    @property
    def VAR(self):
        return self.POWER

    @property
    def UNARY(self):
        return 12

    @property
    def ARITHMATIC_MULT(self):
        return 11

    @property
    def ARITHMATIC_ADD(self):
        return 10

    @property
    def SHIFT(self):
        return 9

    @property
    def UNARY_AND(self):
        return 8

    @property
    def UNARY_OR_XOR(self):
        return 7

    @property
    def COMPARE(self):
        return 6

    @property
    def EQUALITY(self):
        return 5


PRECEDENCE = PrecedenceValues()



from nbtex.core.Ast import makeVar
class ArrowsSymbols:
    def __init__(self):
        pass

    @property
    def Leftarrow(self):
        return makeVar(r"\Leftarrow")

    @property
    def leftharpoonup(self):
        return makeVar(r"\leftharpoonup")

    @property
    def leftharpoondown(self):
        return makeVar(r"\leftharpoondown")

    @property
    def leftleftarrows(self):
        return makeVar(r"\leftleftarrows")

    @property
    def rightarrow(self):
        return makeVar(r"\rightarrow")

    @property
    def Rightarrow(self):
        return makeVar(r"\Rightarrow")

    @property
    def rightharpoonup(self):
        return makeVar(r"\rightharpoonup")

    @property
    def rightharpoondown(self):
        return makeVar(r"\rightharpoondown")

    @property
    def rightrightarrows(self):
        return makeVar(r"\rightrightarrows")

    @property
    def uparrow(self):
        return makeVar(r"\uparrow")

    @property
    def Uparrow(self):
        return makeVar(r"\Uparrow")

    @property
    def upharpoonleft(self):
        return makeVar(r"\upharpoonleft")

    @property
    def upharpoonright(self):
        return makeVar(r"\upharpoonright")

    @property
    def upuparrows(self):
        return makeVar(r"\upuparrows")

    @property
    def downarrow(self):
        return makeVar(r"\downarrow")

    @property
    def Downarrow(self):
        return makeVar(r"\Downarrow")

    @property
    def downharpoonleft(self):
        return makeVar(r"\downharpoonleft")

    @property
    def downharpoonright(self):
        return makeVar(r"\downharpoonright")

    @property
    def downdownarrows(self):
        return makeVar(r"\downdownarrows")

    @property
    def longleftarrow(self):
        return makeVar(r"\longleftarrow")

    @property
    def Longleftarrow(self):
        return makeVar(r"\Longleftarrow")

    @property
    def hookleftarrow(self):
        return makeVar(r"\hookleftarrow")

    @property
    def leftrightharpoons(self):
        return makeVar(r"\leftrightharpoons")

    @property
    def leftrightarrows(self):
        return makeVar(r"\leftrightarrows")

    @property
    def Lleftarrow(self):
        return makeVar(r"\Lleftarrow")

    @property
    def Lsh(self):
        return makeVar(r"\Lsh")

    @property
    def leftarrowtail(self):
        return makeVar(r"\leftarrowtail")

    @property
    def twoheadleftarrow(self):
        return makeVar(r"\twoheadleftarrow")

    @property
    def curvearrowleft(self):
        return makeVar(r"\curvearrowleft")

    @property
    def circlearrowleft(self):
        return makeVar(r"\circlearrowleft")

    @property
    def looparrowleft(self):
        return makeVar(r"\looparrowleft")

    @property
    def longrightarrow(self):
        return makeVar(r"\longrightarrow")

    @property
    def Longrightarrow(self):
        return makeVar(r"\Longrightarrow")

    @property
    def hookrightarrow(self):
        return makeVar(r"\hookrightarrow")

    @property
    def rightleftharpoons(self):
        return makeVar(r"\rightleftharpoons")

    @property
    def rightleftarrows(self):
        return makeVar(r"\rightleftarrows")

    @property
    def Rrightarrow(self):
        return makeVar(r"\Rrightarrow")

    @property
    def Rsh(self):
        return makeVar(r"\Rsh")

    @property
    def rightarrowtail(self):
        return makeVar(r"\rightarrowtail")

    @property
    def twoheadrightarrow(self):
        return makeVar(r"\twoheadrightarrow")

    @property
    def curvearrowright(self):
        return makeVar(r"\curvearrowright")

    @property
    def circlearrowright(self):
        return makeVar(r"\circlearrowright")

    @property
    def looparrowright(self):
        return makeVar(r"\looparrowright")

    @property
    def leftrightarrow(self):
        return makeVar(r"\leftrightarrow")

    @property
    def longleftrightarrow(self):
        return makeVar(r"\longleftrightarrow")

    @property
    def updownarrow(self):
        return makeVar(r"\updownarrow")

    @property
    def nleftarrow(self):
        return makeVar(r"\nleftarrow")

    @property
    def nrightarrow(self):
        return makeVar(r"\nrightarrow")

    @property
    def nleftrightarrow(self):
        return makeVar(r"\nleftrightarrow")

    @property
    def Leftrightarrow(self):
        return makeVar(r"\Leftrightarrow")

    @property
    def Longleftrightarrow(self):
        return makeVar(r"\Longleftrightarrow")

    @property
    def Updownarrow(self):
        return makeVar(r"\Updownarrow")

    @property
    def nLeftarrow(self):
        return makeVar(r"\nLeftarrow")

    @property
    def nRightarrow(self):
        return makeVar(r"\nRightarrow")

    @property
    def nLeftrightarrow(self):
        return makeVar(r"\nLeftrightarrow")

    @property
    def mapsto(self):
        return makeVar(r"\mapsto")

    @property
    def longmapsto(self):
        return makeVar(r"\longmapsto")

    @property
    def multimap(self):
        return makeVar(r"\multimap")

    @property
    def rightsquigarrow(self):
        return makeVar(r"\rightsquigarrow")

    @property
    def leftrightsquigarrow(self):
        return makeVar(r"\leftrightsquigarrow")

    @property
    def swarrow(self):
        return makeVar(r"\swarrow")

    @property
    def searrow(self):
        return makeVar(r"\searrow")

    @property
    def nwarrow(self):
        return makeVar(r"\nwarrow")

    @property
    def nearrow(self):
        return makeVar(r"\nearrow")


Arrow = ArrowsSymbols()


class LatexMatrixFormatter:
    def __init__(self):
        pass

    @staticmethod
    def row_to_latex(row):
        s = ""
        for ele in row:
            s += " " + ele + " & "
        s = s[:-2] + r"\\"
        return s

    @staticmethod
    def matrix_to_latex(matrix):
        s = ""
        for row in matrix:
            s += LatexMatrixFormatter.row_to_latex(row)
        s = s[:-2]
        return s

    @staticmethod
    def create_env(surround):
        env = (
            r"{bmatrix}"
            if surround == "[]"
            else r"{pmatrix}"
            if surround == "()"
            else r"{vmatrix}"
            if surround == "||"
            else r"{Vmatrix}"
            if (surround == "|| ||" or surround == "||||")
            else r"{matrix}"
        )
        begin = r"\begin" + env
        end = r"\end" + env
        return begin, end

    @staticmethod
    def subscript_of_matrix(subscript):
        return "_{" + subscript + "}"

    @staticmethod
    def power_of_matrix(power):
        return "^{" + power + "}"

    @staticmethod
    def empty_power():
        return ""

    @staticmethod
    def empty_subscript():
        return ""



from nbtex.core.Ast import makeVar
class OperatorSymbols:
    def __init__(self):
        pass

    @property
    def plus(self):
        return makeVar("+")

    @property
    def minus(self):
        return makeVar("-")

    @property
    def multiply(self):
        return makeVar("*")

    @property
    def power(self):
        return makeVar("^")

    @property
    def space(self):
        return makeVar("\\hspace{1mm}")

    @property
    def eq(self):
        return makeVar("=")

    @property
    def lt(self):
        return makeVar("<")

    @property
    def leq(self):
        return makeVar(r"\leq")

    @property
    def leqq(self):
        return makeVar(r"\leqq")

    @property
    def leqslant(self):
        return makeVar(r"\leqslant")

    @property
    def gt(self):
        return makeVar(">")

    @property
    def geq(self):
        return makeVar(r"\geq")

    @property
    def geqq(self):
        return makeVar(r"\geqq")

    @property
    def geqslant(self):
        return makeVar(r"\geqslant")

    @property
    def nless(self):
        return makeVar(r"\nless")

    @property
    def nleq(self):
        return makeVar(r"\nleq")

    @property
    def lneq(self):
        return makeVar(r"\lneq")

    @property
    def nleqq(self):
        return makeVar(r"\nleqq")

    @property
    def lneqq(self):
        return makeVar(r"\lneqq")

    @property
    def lvertneqq(self):
        return makeVar(r"\lvertneqq")

    @property
    def nleqslant(self):
        return makeVar(r"\nleqslant")

    @property
    def ngtr(self):
        return makeVar(r"\ngtr")

    @property
    def ngeq(self):
        return makeVar(r"\ngeq")

    @property
    def gneq(self):
        return makeVar(r"\gneq")

    @property
    def ngeqq(self):
        return makeVar(r"\ngeqq")

    @property
    def gneqq(self):
        return makeVar(r"\gneqq")

    @property
    def gvertneqq(self):
        return makeVar(r"\gvertneqq")

    @property
    def ngeqslant(self):
        return makeVar(r"\ngeqslant")

    @property
    def vartriangleleft(self):
        return makeVar(r"\vartriangleleft")

    @property
    def trianglelefteq(self):
        return makeVar(r"\trianglelefteq")

    @property
    def lesssim(self):
        return makeVar(r"\lesssim")

    @property
    def lessapprox(self):
        return makeVar(r"\lessapprox")

    @property
    def prec(self):
        return makeVar(r"\prec")

    @property
    def preceq(self):
        return makeVar(r"\preceq")

    @property
    def precsim(self):
        return makeVar(r"\precsim")

    @property
    def precapprox(self):
        return makeVar(r"\precapprox")

    @property
    def vartriangleright(self):
        return makeVar(r"\vartriangleright")

    @property
    def trianglerighteq(self):
        return makeVar(r"\trianglerighteq")

    @property
    def gtrsim(self):
        return makeVar(r"\gtrsim")

    @property
    def gtrapprox(self):
        return makeVar(r"\gtrapprox")

    @property
    def succ(self):
        return makeVar(r"\succ")

    @property
    def succeq(self):
        return makeVar(r"\succeq")

    @property
    def succsim(self):
        return makeVar(r"\succsim")

    @property
    def succapprox(self):
        return makeVar(r"\succapprox")

    @property
    def ntriangleleft(self):
        return makeVar(r"\ntriangleleft")

    @property
    def ntrianglelefteq(self):
        return makeVar(r"\ntrianglelefteq")

    @property
    def lnsim(self):
        return makeVar(r"\lnsim")

    @property
    def lnapprox(self):
        return makeVar(r"\lnapprox")

    @property
    def nprec(self):
        return makeVar(r"\nprec")

    @property
    def npreceq(self):
        return makeVar(r"\npreceq")

    @property
    def precnsim(self):
        return makeVar(r"\precnsim")

    @property
    def precnapprox(self):
        return makeVar(r"\precnapprox")

    @property
    def ntriangleright(self):
        return makeVar(r"\ntriangleright")

    @property
    def ntrianglerighteq(self):
        return makeVar(r"\ntrianglerighteq")

    @property
    def gnsim(self):
        return makeVar(r"\gnsim")

    @property
    def gnapprox(self):
        return makeVar(r"\gnapprox")

    @property
    def nsucc(self):
        return makeVar(r"\nsucc")

    @property
    def nsucceq(self):
        return makeVar(r"\nsucceq")

    @property
    def succnsim(self):
        return makeVar(r"\succnsim")

    @property
    def succnapprox(self):
        return makeVar(r"\succnapprox")

    @property
    def eqslantless(self):
        return makeVar(r"\eqslantless")

    @property
    def lessgtr(self):
        return makeVar(r"\lessgtr")

    @property
    def lesseqgtr(self):
        return makeVar(r"\lesseqgtr")

    @property
    def lesseqqgtr(self):
        return makeVar(r"\lesseqqgtr")

    @property
    def eqslantgtr(self):
        return makeVar(r"\eqslantgtr")

    @property
    def gtrless(self):
        return makeVar(r"\gtrless")

    @property
    def gtreqless(self):
        return makeVar(r"\gtreqless")

    @property
    def gtreqqless(self):
        return makeVar(r"\gtreqqless")

    @property
    def ll(self):
        return makeVar(r"\ll")

    @property
    def lll(self):
        return makeVar(r"\lll")

    @property
    def lessdot(self):
        return makeVar(r"\lessdot")

    @property
    def preccurlyeq(self):
        return makeVar(r"\preccurlyeq")

    @property
    def curlyeqprec(self):
        return makeVar(r"\curlyeqprec")

    @property
    def gg(self):
        return makeVar(r"\gg")

    @property
    def ggg(self):
        return makeVar(r"\ggg")

    @property
    def gtrdot(self):
        return makeVar(r"\gtrdot")

    @property
    def succcurlyeq(self):
        return makeVar(r"\succcurlyeq")

    @property
    def curlyeqsucc(self):
        return makeVar(r"\curlyeqsucc")

    @property
    def neq(self):
        return makeVar(r"\neq")

    @property
    def equiv(self):
        return makeVar(r"\equiv")

    @property
    def thickapprox(self):
        return makeVar(r"\thickapprox")

    @property
    def approx(self):
        return makeVar(r"\approx")

    @property
    def approxeq(self):
        return makeVar(r"\approxeq")

    @property
    def cong(self):
        return makeVar(r"\cong")

    @property
    def ncong(self):
        return makeVar(r"\ncong")

    @property
    def sim(self):
        return makeVar(r"\sim")

    @property
    def thicksim(self):
        return makeVar(r"\thicksim")

    @property
    def nsim(self):
        return makeVar(r"\nsim")

    @property
    def simeq(self):
        return makeVar(r"\simeq")

    @property
    def backsim(self):
        return makeVar(r"\backsim")

    @property
    def backsimeq(self):
        return makeVar(r"\backsimeq")

    @property
    def eqsim(self):
        return makeVar(r"\eqsim")

    @property
    def doteq(self):
        return makeVar(r"\doteq")

    @property
    def div(self):
        return makeVar(r"\div")

    @property
    def doteqdot(self):
        return makeVar(r"\doteqdot")

    @property
    def fallingdotseq(self):
        return makeVar(r"\fallingdotseq")

    @property
    def risingdotseq(self):
        return makeVar(r"\risingdotseq")

    @property
    def triangleq(self):
        return makeVar(r"\triangleq")

    @property
    def circeq(self):
        return makeVar(r"\circeq")

    @property
    def eqcirc(self):
        return makeVar(r"\eqcirc")

    @property
    def bumpeq(self):
        return makeVar(r"\bumpeq")

    @property
    def Bumpeq(self):
        return makeVar(r"\Bumpeq")

    @property
    def asymp(self):
        return makeVar(r"\asymp")

    @property
    def mid(self):
        return makeVar(r"\mid")

    @property
    def shortmid(self):
        return makeVar(r"\shortmid")

    @property
    def vdash(self):
        return makeVar(r"\vdash")

    @property
    def dashv(self):
        return makeVar(r"\dashv")

    @property
    def Vdash(self):
        return makeVar(r"\Vdash")

    @property
    def parallel(self):
        return makeVar(r"\parallel")

    @property
    def shortparallel(self):
        return makeVar(r"\shortparallel")

    @property
    def vDash(self):
        return makeVar(r"\vDash")

    @property
    def Vvdash(self):
        return makeVar(r"\Vvdash")

    @property
    def models(self):
        return makeVar(r"\models")

    @property
    def nmid(self):
        return makeVar(r"\nmid")

    @property
    def nshortmid(self):
        return makeVar(r"\nshortmid")

    @property
    def nvdash(self):
        return makeVar(r"\nvdash")

    @property
    def nVdash(self):
        return makeVar(r"\nVdash")

    @property
    def nparallel(self):
        return makeVar(r"\nparallel")

    @property
    def nshortparallel(self):
        return makeVar(r"\nshortparallel")

    @property
    def nvDash(self):
        return makeVar(r"\nvDash")

    @property
    def nVDash(self):
        return makeVar(r"\nVDash")


Operator = OperatorSymbols()


from .LatexBasicFormatter import LatexBasicFormatter
from .LatexDerivativesFormatter import LatexDerivativesFormatter
from .LatexSeriesFormatter import LatexSeriesFormatter
from .LatexMatrixFormatter import LatexMatrixFormatter



"Methods for extracting properties from: https://en.wikipedia.org/wiki/Wikipedia:LaTeX_symbols"
import re


def getNameSym(pasta):
    lst = []
    for l in pasta.split("\n"):
        r = re.search("{.*}", l)
        if r is not None:
            p = r.group()[1:-1]
            ps = p.split(" ")
            if len(ps) > 1 and ps[1] != "\\":
                lst.append(ps[1])
    return [(arg[1:], arg) for arg in lst]


def getDefs(l):
    for name, ap in l:
        name = ap[1:]
        print(
            f"""@property
def {name}(self):
    return Var(r'{ap}')"""
        )


def getProps(cl):
    l = []
    for v in vars(cl):
        if isinstance(vars(cl)[v], property):
            l.append(v)
    return l


def getWrappedProps(cl, objName):
    l = getProps(cl)
    s = ""
    for name in l:
        s += f"{objName}.{name} | "
    sx = f"wrap({s[:-2]})"
    s = "{{" + f"wrap({s[:-2]})" + "}}"
    return sx



from nbtex.core.operators.Operator import (
    BasicOperator,
    InvertibleInfixOperator,
    InvertibleOperator,
    infix_combine,
)
from nbtex.core.Precedence import PRECEDENCE
from nbtex.LatexInterface import Operator, LatexBasicFormatter

Plus = InvertibleInfixOperator(PRECEDENCE.ARITHMATIC_ADD, Operator.plus, Operator.minus)
Minus = ~Plus
Equal = InvertibleInfixOperator(PRECEDENCE.EQUALITY, Operator.eq, Operator.neq)
LessThan = InvertibleInfixOperator(PRECEDENCE.COMPARE, Operator.lt, Operator.nless)
LessThanEqual = InvertibleInfixOperator(PRECEDENCE.COMPARE, Operator.leq, Operator.nleq)
GreaterThan = InvertibleInfixOperator(PRECEDENCE.COMPARE, Operator.gt, Operator.ngtr)
GreaterThanEqual = InvertibleInfixOperator(
    PRECEDENCE.COMPARE, Operator.geq, Operator.ngeq
)
Space = InvertibleInfixOperator(PRECEDENCE.EQUALITY, Operator.space)
Newline = InvertibleInfixOperator(PRECEDENCE.EQUALITY, Operator.newline)

Negate = InvertibleOperator(
    PRECEDENCE.POWER, LatexBasicFormatter.negate, LatexBasicFormatter.unnegate
)

mult_combine = infix_combine(Operator.multiply)
frac_combine = LatexBasicFormatter.fraction

Multiply = InvertibleOperator(PRECEDENCE.ARITHMATIC_MULT, mult_combine, frac_combine)
Fraction = ~Multiply

power_combine = infix_combine(Operator.power)

Power = InvertibleOperator(PRECEDENCE.POWER, power_combine, LatexBasicFormatter.root)
Root = ~Power



import unittest

from nbtex.core.operators import (
    InvertibleInfixOperator,
    InvertibleOperator,
    BasicOperator,
)


def _push_combine(string):
    return string + "1"


def _pop_combine(string):
    return string[:-1] if len(string) > 0 else string


class TestCoreOperators(unittest.TestCase):
    def test_basic_operator(self):
        basic_operator = BasicOperator(0, _push_combine)
        output = basic_operator("x2")
        self.assertEqual(output, "x21")

    def test_invertible_operator(self):
        invertible_operator = InvertibleOperator(0, _push_combine, _pop_combine)
        after_combine = invertible_operator("x")
        self.assertEqual(after_combine, "x1")
        after_inv_combine = (~invertible_operator)(after_combine)
        self.assertEqual(after_inv_combine, "x")

    def test_invertible_operator_without_invert(self):
        invertible_operator = InvertibleOperator(0, _push_combine)
        after_combine = invertible_operator("x")
        after_inv_combine = (~invertible_operator)("x")
        self.assertEqual(after_combine, "x1")
        self.assertEqual(after_inv_combine, "x1")

    def test_precedence(self):
        operator = BasicOperator(0, _push_combine)
        invertible_operator = InvertibleOperator(0, _push_combine)
        invertible_infix_operator = InvertibleInfixOperator(0, "+")
        self.assertEqual(operator.precedence, 0)
        self.assertEqual(invertible_operator.precedence, 0)
        self.assertEqual(invertible_infix_operator.precedence, 0)

    def test_invertible_infix_operator(self):
        operator = InvertibleInfixOperator(0, "+", "-")
        after_combine = operator("1", "2")
        after_invert_combine = (~operator)("1", "2")
        self.assertEqual(after_combine, "1 + 2")
        self.assertEqual(after_invert_combine, "1 - 2")

    def test_inveritble_infix_without_invert(self):
        operator = InvertibleInfixOperator(0, "+")
        after_combine = operator("1", "2")
        after_invert_combine = (~operator)("1", "2")
        self.assertEqual(after_combine, "1 + 2")
        self.assertEqual(after_invert_combine, "1 + 2")


if __name__ == "__main__":
    unittest.main()



import unittest
from nbtex.LatexInterface.LatexFormatters import LatexBasicFormatter

class TestLatexBasicFormatter(unittest.TestCase):
    def test_bold_italics_underline(self):
        bold_string = LatexBasicFormatter.bold("1")
        italics_string = LatexBasicFormatter.italics("1")
        underline_string = LatexBasicFormatter.underline("1")
        self.assertEqual(bold_string, r"\textbf{1}")
        self.assertEqual(italics_string, r"\textit{1}")
        self.assertEqual(underline_string, r"\underline{1}")

    def test_binary_operation_output(self):
        add_output = LatexBasicFormatter.binary_operation_output("+", "1", "2")
        sub_output = LatexBasicFormatter.binary_operation_output("-", "1", "2")
        self.assertEqual(add_output, "1 + 2")
        self.assertEqual(sub_output, '1 - 2')

    def test_surround_with_parens(self):
        output = LatexBasicFormatter.surround_with_parens("1")
        self.assertEqual(output, r"\left(1\right)")

    def test_negate_unnegate(self):
        negative = LatexBasicFormatter.negate('123')
        still_negative = LatexBasicFormatter.negate('-123')
        positive = LatexBasicFormatter.unnegate('-123')
        still_positive = LatexBasicFormatter.unnegate('123')
        self.assertEqual(negative, '-123')
        self.assertEqual(still_negative, '-123')
        self.assertEqual(positive, '123')
        self.assertEqual(still_positive, '123')

    def test_fraction(self):
        fraction = LatexBasicFormatter.fraction('1', '2')
        self.assertEqual(fraction, r'\frac{1}{2}')
    
    def test_subscript_superscript(self):
        subscript = LatexBasicFormatter.subscript('1')
        superscript = LatexBasicFormatter.superscript('1')
        self.assertEqual(subscript, '_1')
        self.assertEqual(superscript, '^1')

    def test_root(self):
        nthRoot = LatexBasicFormatter.root('1', 'n')
        self.assertEqual(nthRoot, r'\sqrt[\leftroot{-1}\uproot{1}n]{1}')

    def test_function_def(self):
        noarg = LatexBasicFormatter.function_definition('g')
        monoarg = LatexBasicFormatter.function_definition('g', 'x')
        twoArg = LatexBasicFormatter.function_definition('g', 'x', 'y')
        self.assertEqual(noarg, 'g()')
        self.assertEqual(monoarg, 'g(x)')
        self.assertEqual(twoArg, 'g(x, y)')

    def test_inverse_function_def(self):
        noarg = LatexBasicFormatter.inverse_function_definition('g')
        monoarg = LatexBasicFormatter.inverse_function_definition('g', 'x')
        twoArg = LatexBasicFormatter.inverse_function_definition('g', 'x', 'y')
        self.assertEqual(noarg, 'g^{-1}()')
        self.assertEqual(monoarg, 'g^{-1}(x)')
        self.assertEqual(twoArg, 'g^{-1}(x, y)')


class LatexBasicFormatter:
    def __init__(self):
        pass

    @staticmethod
    def bold(string):
        return "\\textbf{" + string + "}"

    @staticmethod
    def italics(string):
        return "\\textit{" + string + "}"

    @staticmethod
    def underline(string):
        return "\\underline{" + string + "}"

    @staticmethod
    def binary_operation_output(binary_operator, a1, a2):
        return f"{a1} {binary_operator} {a2}"

    @staticmethod
    def surround_with_parens(string):
        return f"\\left({string}\\right)"

    @staticmethod
    def negate(string):
        if string[0] == "-":
            return string
        else:
            return f"-{string}"

    @staticmethod
    def unnegate(string):
        return string[1:] if (string[0] == "-") else string

    @staticmethod
    def fraction(num, den):
        num, den = LatexBasicFormatter.surround_all_with_braces(num, den)
        return f"\\frac{num}{den}"

    @staticmethod
    def subscript(string):
        return f"_{string}"

    @staticmethod
    def superscript(string):
        return f"^{string}"

    @staticmethod
    def surround_with_braces(string):
        return "{" + string + "}"

    @staticmethod
    def surround_all_with_braces(*args):
        if len(args) == 1:
            return LatexBasicFormatter.surround_with_braces(args[0])
        else:
            return [LatexBasicFormatter.surround_with_braces(arg) for arg in args]

    @staticmethod
    def root(num, nth):
        return r"\sqrt[\leftroot{-1}\uproot{1}" + nth + r"]{" + num + r"}"

    @staticmethod
    def function_definition(f, *args):
        if len(args) == 0:
            return f"{f}()"
        s = ""
        for arg in args:
            s += f"{arg}, "
        s = s[:-2]
        return f"{f}({s})"

    @staticmethod
    def inverse_function_definition(f, *args):
        _1 = LatexBasicFormatter.surround_with_braces("-1")
        return LatexBasicFormatter.function_definition(f"{f}^{_1}", *args)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="nbtex",
    version="1.0.6",
    author="Hariom Narang",
    author_email="hariom2711@gmail.com",
    packages=find_packages(),
    url="https://github.com/narang99/nbtex",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
    description="A light-weight wrapper around LaTeX for mathematical symbols for Jupyter Notebooks",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from nbtex.core.Ast import makeVar
class GreekSymbols:
    def __init__(self):
        pass

    @property
    def alpha(self):
        return makeVar(r"\alpha")

    @property
    def beta(self):
        return makeVar(r"\beta")

    @property
    def gamma(self):
        return makeVar(r"\gamma")

    @property
    def delta(self):
        return makeVar(r"\delta")

    @property
    def epsilon(self):
        return makeVar(r"\epsilon")

    @property
    def zeta(self):
        return makeVar(r"\zeta")

    @property
    def eta(self):
        return makeVar(r"\eta")

    @property
    def theta(self):
        return makeVar(r"\theta")

    @property
    def iota(self):
        return makeVar(r"\iota")

    @property
    def kappa(self):
        return makeVar(r"\kappa")

    @property
    def lambd(self):
        return makeVar(r"\lambda")

    @property
    def mu(self):
        return makeVar(r"\mu")

    @property
    def nu(self):
        return makeVar(r"\nu")

    @property
    def xi(self):
        return makeVar(r"\xi")

    @property
    def mathrm(self):
        return makeVar(r"\mathrm")

    @property
    def pi(self):
        return makeVar(r"\pi")

    @property
    def rho(self):
        return makeVar(r"\rho")

    @property
    def sigma(self):
        return makeVar(r"\sigma")

    @property
    def tau(self):
        return makeVar(r"\tau")

    @property
    def upsilon(self):
        return makeVar(r"\upsilon")

    @property
    def phi(self):
        return makeVar(r"\phi")

    @property
    def chi(self):
        return makeVar(r"\chi")

    @property
    def psi(self):
        return makeVar(r"\psi")

    @property
    def omega(self):
        return makeVar(r"\omega")

    @property
    def varepsilon(self):
        return makeVar(r"\varepsilon")

    @property
    def vartheta(self):
        return makeVar(r"\vartheta")

    @property
    def varkappa(self):
        return makeVar(r"\varkappa")

    @property
    def varpi(self):
        return makeVar(r"\varpi")

    @property
    def varrho(self):
        return makeVar(r"\varrho")

    @property
    def varphi(self):
        return makeVar(r"\varphi")

    @property
    def varsigma(self):
        return makeVar(r"\varsigma")

    @property
    def Gamma(self):
        return makeVar(r"\Gamma")

    @property
    def Delta(self):
        return makeVar(r"\Delta")

    @property
    def Theta(self):
        return makeVar(r"\Theta")

    @property
    def Lambda(self):
        return makeVar(r"\Lambda")

    @property
    def Upsilon(self):
        return makeVar(r"\Upsilon")

    @property
    def Xi(self):
        return makeVar(r"\Xi")

    @property
    def Phi(self):
        return makeVar(r"\Phi")

    @property
    def Pi(self):
        return makeVar(r"\Pi")

    @property
    def Psi(self):
        return makeVar(r"\Psi")

    @property
    def Sigma(self):
        return makeVar(r"\Sigma")

    @property
    def Omega(self):
        return makeVar(r"\Omega")

    @property
    def digamma(self):
        return makeVar(r"\digamma")

    @property
    def aleph(self):
        return makeVar(r"\aleph")

    @property
    def beth(self):
        return makeVar(r"\beth")

    @property
    def gimel(self):
        return makeVar(r"\gimel")

    @property
    def daleth(self):
        return makeVar(r"\daleth")

    @property
    def forall(self):
        return makeVar(r"\forall")

    @property
    def exists(self):
        return makeVar(r"\exists")

    @property
    def nexists(self):
        return makeVar(r"\nexists")

    @property
    def Finv(self):
        return makeVar(r"\Finv")

    @property
    def Game(self):
        return makeVar(r"\Game")

    @property
    def backepsilon(self):
        return makeVar(r"\backepsilon")

    @property
    def imath(self):
        return makeVar(r"\imath")

    @property
    def jmath(self):
        return makeVar(r"\jmath")

    @property
    def ell(self):
        return makeVar(r"\ell")

    @property
    def Bbbk(self):
        return makeVar(r"\Bbbk")

    @property
    def amalg(self):
        return makeVar(r"\amalg")

    @property
    def nabla(self):
        return makeVar(r"\nabla")

    @property
    def mho(self):
        return makeVar(r"\mho")

    @property
    def partial(self):
        return makeVar(r"\partial")

    @property
    def eth(self):
        return makeVar(r"\eth")

    @property
    def hbar(self):
        return makeVar(r"\hbar")

    @property
    def hslash(self):
        return makeVar(r"\hslash")

    @property
    def Im(self):
        return makeVar(r"\Im")

    @property
    def Re(self):
        return makeVar(r"\Re")

    @property
    def wp(self):
        return makeVar(r"\wp")

    @property
    def emptyset(self):
        return makeVar(r"\emptyset")

    @property
    def varnothing(self):
        return makeVar(r"\varnothing")

    @property
    def complement(self):
        return makeVar(r"\complement")

    @property
    def circledS(self):
        return makeVar(r"\circledS")


"default symbol lookup class, can change name later maybe"
Greek = GreekSymbols()


from nbtex.core.operators.Operator import InvertibleOperator
from nbtex.LatexInterface.LatexFormatters import LatexSeriesFormatter
from nbtex.core.Precedence import PRECEDENCE

Summation = InvertibleOperator(PRECEDENCE.POWER, LatexSeriesFormatter.summation)

Product = InvertibleOperator(PRECEDENCE.POWER, LatexSeriesFormatter.product)



from nbtex.LatexInterface.LatexFormatters import LatexBasicFormatter
from functools import partial


class BasicOperator:
    def __init__(self, precedence, combine):
        self._precedence, self._combine = precedence, combine

    def __call__(self, *args):
        return self._combine(*args)

    @property
    def precedence(self):
        return self._precedence


class InvertibleOperator(BasicOperator):
    def __init__(self, precedence, combine, invert_combine=None):
        super().__init__(precedence, combine)
        self._invert_combine = invert_combine

    def __call__(self, *args):
        return super().__call__(*args)

    def __invert__(self):
        if self._invert_combine is not None:
            return InvertibleOperator(
                self.precedence, self._invert_combine, self._combine
            )
        else:
            return self


def infix_combine(op):
    return partial(LatexBasicFormatter.binary_operation_output, op)


class InvertibleInfixOperator(InvertibleOperator):
    def __init__(self, precedence, op, invert_op=None):
        c, invc = infix_combine(op), infix_combine(invert_op)
        if invert_op is None:
            super().__init__(precedence, c)
        else:
            super().__init__(precedence, c, invc)



import unittest
from nbtex.LatexInterface.LatexFormatters import LatexDerivativesFormatter, LatexBasicFormatter

class TestLatexDerivativesFormatter(unittest.TestCase):
    def test_partial(self):
        partial = LatexDerivativesFormatter.partial('x')
        self.assertEqual(LatexBasicFormatter.fraction(r'\partial', r'\partial x'), partial)

    def test_partial_of_one_word(self):
        partial_of_one_word = LatexDerivativesFormatter.partial_of('xy', 'x')
        partial_of_one_word_power = LatexDerivativesFormatter.partial_of('x^y', 'x')
        partial_of_one_word_subscript = LatexDerivativesFormatter.partial_of('x_y','x')
        partial_of_one_word_power_subscript = LatexDerivativesFormatter.partial_of('x_y^z', 'x')
        self.assertEqual(partial_of_one_word, LatexBasicFormatter.fraction(r'\partial xy', r'\partial x'))
        self.assertEqual(partial_of_one_word_power, LatexBasicFormatter.fraction(r'\partial x^y', r'\partial x'))
        self.assertEqual(partial_of_one_word_subscript, LatexBasicFormatter.fraction(r'\partial x_y', r'\partial x'))
        self.assertEqual(partial_of_one_word_power_subscript, LatexBasicFormatter.fraction(r'\partial x_y^z', r'\partial x'))
    
    def test_partial_of_multi_word(self):
        partial = LatexDerivativesFormatter.partial_of('x+1', 'x')
        par = LatexDerivativesFormatter.partial('x')
        self.assertEqual(partial, par + r' \left(x+1\right)')

    def test_derivative(self):
        derivative = LatexDerivativesFormatter.derivative('x')
        self.assertEqual(LatexBasicFormatter.fraction(r'd', r'dx'), derivative)

    def test_derivative_of_one_word(self):
        derivative_of_one_word = LatexDerivativesFormatter.derivative_of('xy', 'x')
        derivative_of_one_word_power = LatexDerivativesFormatter.derivative_of('x^y', 'x')
        derivative_of_one_word_subscript = LatexDerivativesFormatter.derivative_of('x_y','x')
        derivative_of_one_word_power_subscript = LatexDerivativesFormatter.derivative_of('x_y^z', 'x')
        self.assertEqual(derivative_of_one_word, LatexBasicFormatter.fraction(r'dxy', r'dx'))
        self.assertEqual(derivative_of_one_word_power, LatexBasicFormatter.fraction(r'dx^y', r'dx'))
        self.assertEqual(derivative_of_one_word_subscript, LatexBasicFormatter.fraction(r'dx_y', r'dx'))
        self.assertEqual(derivative_of_one_word_power_subscript, LatexBasicFormatter.fraction(r'dx_y^z', r'dx'))
    
    def test_derivative_of_multi_word(self):
        derivative = LatexDerivativesFormatter.derivative_of('x+1', 'x')
        der = LatexDerivativesFormatter.derivative('x')
        self.assertEqual(derivative, der + r' \left(x+1\right)')



from nbtex.core.operators.Operator import InvertibleOperator
from nbtex.LatexInterface.LatexFormatters import (
    LatexSeriesFormatter,
    LatexDerivativesFormatter,
)
from nbtex.core.Precedence import PRECEDENCE


def inv_integral(function, wrt, lower="", upper=""):
    return LatexDerivativesFormatter.derivative_of(function, wrt)


Integral = InvertibleOperator(
    PRECEDENCE.POWER, LatexSeriesFormatter.integral, inv_integral
)
Derivative = ~Integral


def inv_partial(function, wrt, lower="", upper=""):
    return LatexDerivativesFormatter.partial_of(function, wrt)


PartialIntegral = InvertibleOperator(
    PRECEDENCE.POWER, LatexSeriesFormatter.partial_integral, inv_partial
)
PartialDerivative = ~PartialIntegral



from nbtex.core.Ast import Var


class Dots(Var):
    def __init__(self, direc="h"):
        super().__init__(direc + "dots")
        self.direc = direc

    def build(self):
        return (
            r"\cdots"
            if (self.direc == "h" or self.direc == "c")
            else r"\vdots"
            if (self.direc == "v")
            else r"\ddots"
        )

    def clone(self):
        return Dots(self.direc)


import unittest
from nbtex.core.Ast import Var, makeVar
import nbtex.core.operators.Arithmatic as A


class TestAst(unittest.TestCase):
    def test_var_init(self):
        a = Var("a")
        self.assertEqual(a.build(), "a")
        self.assertIsInstance(a, Var)

    def test_pass_single_in_makeVar(self):
        one_var = makeVar("o")
        self.assertIsInstance(one_var, Var)

    def test_pass_multiple_in_makeVar(self):
        a, b, _1 = makeVar("a", "b", 1)
        self.assertIsInstance(a, Var)
        self.assertIsInstance(b, Var)
        self.assertIsInstance(_1, Var)
        lst = makeVar("a", "b")
        self.assertIsInstance(lst, list)

    def test_pass_Var_in_makeVar(self):
        a = makeVar("a")
        na = makeVar(a)
        self.assertEqual(a.build(), na.build())
        self.assertIsNot(a, na)

    def test_basic_aritmatic(self):
        a, b = makeVar("a", "b")
        c = a + b
        self.assertIsNot(c, a)
        self.assertIsNot(c, b)
        self.assertEqual(c.build(), A.Plus("a", "b"))
        c = a - b
        self.assertEqual(c.build(), A.Minus("a", "b"))
        c = a * b
        self.assertEqual(c.build(), A.Multiply("a", "b"))
        c = a / b
        self.assertEqual(c.build(), A.Fraction("a", "b"))
        c = a ** b
        self.assertEqual(c.build(), A.Power("a", "b"))
        c = a.root(b)
        self.assertEqual(c.build(), A.Root('a','b'))

    def test_comparison_operators(self):
        a, b = makeVar("a", "b")
        c = a < b
        self.assertEqual(c.build(), A.LessThan("a", "b"))
        c = a > b
        self.assertEqual(c.build(), A.GreaterThan("a", "b"))
        c = a <= b
        self.assertEqual(c.build(), A.LessThanEqual("a", "b"))
        c = a >= b
        self.assertEqual(c.build(), A.GreaterThanEqual("a", "b"))
        c = a.equals(b)
        self.assertEqual(c.build(), A.Equal("a", "b"))
        c = a.not_equals(b)
        self.assertEqual(c.build(), (~A.Equal)("a", "b"))

    def test_inverted_basic_arithmatic(self):
        a, b = makeVar("a", "b")
        c = a + b
        self.assertIsNot(c, a)
        self.assertIsNot(c, b)
        self.assertEqual((~c).build(), A.Minus("a", "b"))
        c = a - b
        self.assertEqual((~c).build(), A.Plus("a", "b"))
        c = a * b
        self.assertEqual((~c).build(), A.Fraction("a", "b"))
        c = a / b
        self.assertEqual((~c).build(), A.Multiply("a", "b"))
        c = a ** b
        self.assertEqual((~c).build(), A.Root("a", "b"))
        c = a.root(b)
        self.assertEqual((~c).build(), A.Power('a','b'))

    def test_inverted_comparison_operators(self):
        a, b = makeVar("a", "b")
        c = a < b
        self.assertEqual((~c).build(), (~A.LessThan)("a", "b"))
        c = a > b
        self.assertEqual((~c).build(), (~A.GreaterThan)("a", "b"))
        c = a <= b
        self.assertEqual((~c).build(), (~A.LessThanEqual)("a", "b"))
        c = a >= b
        self.assertEqual((~c).build(), (~A.GreaterThanEqual)("a", "b"))
        c = a.equals(b)
        self.assertEqual((~c).build(), (~A.Equal)("a", "b"))




from nbtex.core.Ast import Var
from nbtex.helpers import varArgFunc


class Vector(Var):
    def __init__(self, name):
        super().__init__(r"\vec{\mathbf{" + name + "}}")


def makeVector(*args):
    return varArgFunc(
        lambda arg: arg if isinstance(arg, Vector) else Vector(str(arg)), *args
    )



import unittest
from nbtex.LatexInterface.LatexFormatters import LatexSeriesFormatter

class TestLatexSeriesFormatter(unittest.TestCase):
    def test_summation(self):
        summation_just_wrt = LatexSeriesFormatter.summation('x+1', 'x')
        summation_wrt_lower = LatexSeriesFormatter.summation('x+1', 'x', '1')
        summation_wrt_lower_upper = LatexSeriesFormatter.summation('x+1', 'x', '1', '10')
        self.assertEqual(summation_just_wrt, r'\sum_{x} {x+1}') 
        self.assertEqual(summation_wrt_lower, r'\sum_{x=1} {x+1}')
        self.assertEqual(summation_wrt_lower_upper, r'\sum_{x=1}^{10} {x+1}')

    def test_product(self):
        product_just_wrt = LatexSeriesFormatter.product('x+1', 'x')
        product_wrt_lower = LatexSeriesFormatter.product('x+1', 'x', '1')
        product_wrt_lower_upper = LatexSeriesFormatter.product('x+1', 'x', '1', '10')
        self.assertEqual(product_just_wrt, r'\prod_{x} {x+1}') 
        self.assertEqual(product_wrt_lower, r'\prod_{x=1} {x+1}')
        self.assertEqual(product_wrt_lower_upper, r'\prod_{x=1}^{10} {x+1}')
    
    def test_integral(self):
        int_just_wrt = LatexSeriesFormatter.integral('x+1', 'x')
        int_wrt_lower = LatexSeriesFormatter.integral('x+1', 'x', '1')
        int_wrt_lower_upper = LatexSeriesFormatter.integral('x+1', 'x', '1', '10')
        self.assertEqual(int_just_wrt, r'\int {x+1} \hspace{1mm} {dx}') 
        self.assertEqual(int_wrt_lower, r'\int_{1} {x+1} \hspace{1mm} {dx}')
        self.assertEqual(int_wrt_lower_upper, r'\int_{1}^{10} {x+1} \hspace{1mm} {dx}')

    def test_partial_integral(self):
        partial_just_wrt = LatexSeriesFormatter.partial_integral('x+1', 'x')
        partial_wrt_lower = LatexSeriesFormatter.partial_integral('x+1', 'x', '1')
        partial_wrt_lower_upper = LatexSeriesFormatter.partial_integral('x+1', 'x', '1', '10')
        self.assertEqual(partial_just_wrt, r'\int {x+1} \hspace{1mm} {\partial x}') 
        self.assertEqual(partial_wrt_lower, r'\int_{1} {x+1} \hspace{1mm} {\partial x}')
        self.assertEqual(partial_wrt_lower_upper, r'\int_{1}^{10} {x+1} \hspace{1mm} {\partial x}')